from typing import Any, Optional, Iterable

from AnyQt.QtWidgets import (
    QListView, QSizePolicy, QStyle, QStyleOptionViewItem, QWidget
)
from AnyQt.QtCore import Qt, QSize, QModelIndex


class LinearIconView(QListView):
    """
    An list view (in QListView.IconMode) with no item wrapping.

    Suitable for displaying large(ish) icons with text in a single row/column.
    """
    def __init__(self, parent=None, iconSize=QSize(120, 80), **kwargs):
        # type: (Optional[QWidget], QSize, Any)-> None
        super().__init__(parent, **kwargs)
        self.setViewMode(QListView.IconMode)
        self.setWrapping(False)
        self.setWordWrap(True)

        self.setSelectionMode(QListView.SingleSelection)
        self.setEditTriggers(QListView.NoEditTriggers)
        self.setMovement(QListView.Static)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setSizePolicy(QSizePolicy.Expanding,
                           QSizePolicy.Fixed)
        self.setIconSize(iconSize)

    def sizeHint(self):
        # type: () -> QSize
        """
        Reimplemented.

        Provide sensible size hint based on the view's contents.
        """
        flow = self.flow()

        if self.model() is None or not self.model().rowCount():
            style = self.style()
            opt = self.viewOptions()
            opt.features = QStyleOptionViewItem.ViewItemFeature(
                opt.features |
                QStyleOptionViewItem.HasDecoration |
                QStyleOptionViewItem.HasDisplay |
                QStyleOptionViewItem.WrapText
            )
            opt.text = "X" * 12 + "\nX"
            sh = style.sizeFromContents(
                QStyle.CT_ItemViewItem, opt, QSize(), self)
        else:
            # Sample the first 20 items for a size hint. The objective is to
            # get a representative height due to the word wrapping
            model = self.model()
            samplesize = min(20, model.rowCount())
            shs = [self.sizeHintForIndex(model.index(i, 0))
                   for i in range(samplesize)]
            if flow == QListView.TopToBottom:
                sh = QSize(max(s.width() for s in shs), 200)
            else:
                sh = QSize(200, max(s.height() for s in shs))

        margins = self.contentsMargins()
        if flow == QListView.TopToBottom:
            sh = sh + QSize(margins.left() + margins.right(), 0)
        else:
            sh = sh + QSize(0, margins.top() + margins.bottom())

        if flow == QListView.TopToBottom and \
                self.verticalScrollBarPolicy() != Qt.ScrollBarAlwaysOff:
            ssh = self.verticalScrollBar().sizeHint()
            return QSize(sh.width() + ssh.width(), sh.height())
        elif self.flow() == QListView.LeftToRight and \
                self.horizontalScrollBarPolicy() != Qt.ScrollBarAlwaysOff:
            ssh = self.horizontalScrollBar().sizeHint()
            return QSize(sh.width(), sh.height() + ssh.height())
        else:
            return sh

    def updateGeometries(self):
        # type: () -> None
        """Reimplemented"""
        super().updateGeometries()
        self.updateGeometry()

    def dataChanged(self, topLeft, bottomRight, roles=()):
        # type: (QModelIndex, QModelIndex, Iterable[int]) -> None
        """Reimplemented"""
        super().dataChanged(topLeft, bottomRight, roles)
        self.updateGeometry()



import unittest
from collections import Counter

from AnyQt.QtCore import QEvent
from AnyQt.QtWidgets import QWidget, QApplication, QAction
from AnyQt.QtTest import QSignalSpy

from orangecanvas.scheme import (
    Scheme, NodeEvent, SchemeLink, LinkEvent, WorkflowEvent
)
from orangecanvas.scheme.widgetmanager import WidgetManager
from orangecanvas.registry import tests as registry_tests
from orangecanvas.scheme.tests import EventSpy


class TestingWidgetManager(WidgetManager):
    def create_widget_for_node(self, node):
        return QWidget()

    def delete_widget_for_node(self, node, widget):
        widget.deleteLater()

    def save_widget_geometry(self, node, widget):
        return widget.saveGeometry()

    def restore_widget_geometry(self, node, widget, state):
        return widget.restoreGeometry(state)


class TestWidgetManager(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        app = QApplication.instance()
        if app is None:
            app = QApplication([])
        cls.app = app

    @classmethod
    def tearDownClass(cls):
        cls.app = None

    def setUp(self):
        reg = registry_tests.small_testing_registry()
        scheme = Scheme()
        zero = scheme.new_node(reg.widget("zero"))
        one = scheme.new_node(reg.widget("one"))
        add = scheme.new_node(reg.widget("add"))
        scheme.new_link(zero, "value", add, "left")
        scheme.new_link(one, "value", add, "right")

        self.scheme = scheme

    def test_create_immediate(self):
        wm = TestingWidgetManager()
        wm.set_creation_policy(TestingWidgetManager.Immediate)
        spy = QSignalSpy(wm.widget_for_node_added)
        wm.set_workflow(self.scheme)
        nodes = self.scheme.nodes
        self.assertEqual(len(spy), 3)
        self.assertSetEqual({n for n, _ in spy}, set(nodes))
        spy = QSignalSpy(wm.widget_for_node_removed)
        self.scheme.clear()
        self.assertEqual(len(spy), 3)
        self.assertSetEqual({n for n, _ in spy}, set(nodes))

    def test_create_normal(self):
        workflow = self.scheme
        nodes = workflow.nodes
        wm = TestingWidgetManager()
        wm.set_creation_policy(TestingWidgetManager.Normal)
        spy = QSignalSpy(wm.widget_for_node_added)
        wm.set_workflow(workflow)
        self.assertEqual(len(spy), 0)

        w = wm.widget_for_node(nodes[0])
        self.assertEqual(list(spy), [[nodes[0], w]])

        w = wm.widget_for_node(nodes[2])
        self.assertEqual(list(spy)[1:], [[nodes[2], w]])

        spy = QSignalSpy(wm.widget_for_node_added)
        self.assertTrue(spy.wait())
        w = wm.widget_for_node(nodes[1])
        self.assertEqual(list(spy), [[nodes[1], w]])

        spy = QSignalSpy(wm.widget_for_node_removed)
        workflow.clear()
        self.assertEqual(len(spy), 3)
        self.assertSetEqual({n for n, _ in spy}, set(nodes))

    def test_create_on_demand(self):
        workflow = self.scheme
        nodes = workflow.nodes
        wm = TestingWidgetManager()
        wm.set_creation_policy(WidgetManager.OnDemand)
        spy = QSignalSpy(wm.widget_for_node_added)
        wm.set_workflow(workflow)
        self.assertEqual(len(spy), 0)
        self.assertFalse(spy.wait(30))
        self.assertEqual(len(spy), 0)
        w = wm.widget_for_node(nodes[0])
        self.assertEqual(list(spy), [[nodes[0], w]])
        # transition to normal
        spy = QSignalSpy(wm.widget_for_node_added)
        wm.set_creation_policy(WidgetManager.Normal)
        self.assertTrue(spy.wait())
        self.assertEqual(spy[0][0], nodes[1])

    def test_mappings(self):
        workflow = self.scheme
        nodes = workflow.nodes
        wm = TestingWidgetManager()
        wm.set_workflow(workflow)
        w = wm.widget_for_node(nodes[0])
        n = wm.node_for_widget(w)
        self.assertIs(n, nodes[0])

    def test_save_geometry(self):
        workflow = self.scheme
        nodes = workflow.nodes
        wm = TestingWidgetManager()
        wm.set_workflow(workflow)
        n = nodes[0]
        w = wm.widget_for_node(n)
        state = wm.save_widget_geometry(n, w)
        self.assertTrue(wm.restore_widget_geometry(n, w, state))
        wm.activate_widget_for_node(n, w)
        state = wm.save_window_state()

        self.assertEqual(len(state), 1)
        self.assertIs(state[0][0], n)
        self.assertEqual(state[0][1], wm.save_widget_geometry(n, w))
        QApplication.sendEvent(
            nodes[1], NodeEvent(NodeEvent.NodeActivateRequest, nodes[1])
        )
        wm.raise_widgets_to_front()

        wm.restore_window_state(state)

    def test_set_model(self):
        workflow = self.scheme
        wm = TestingWidgetManager()
        wm.set_workflow(workflow)
        wm.set_workflow(workflow)
        wm.set_creation_policy(WidgetManager.Immediate)
        wm.set_workflow(Scheme())

    def test_event_dispatch(self):
        workflow = self.scheme
        nodes = workflow.nodes
        links = workflow.links

        class Widget(QWidget):
            def __init__(self, *a):
                self._evt = []
                super().__init__(*a)

            def event(self, event):
                # record all event types
                self._evt.append(event.type())
                return super().event(event)

        class WidgetManager(TestingWidgetManager):
            def create_widget_for_node(self, node):
                w = Widget()
                w._evt = []
                return w

        wm = WidgetManager()
        wm.set_creation_policy(WidgetManager.OnDemand)
        wm.set_workflow(workflow)
        n1, n2, n3 = nodes[:3]
        l1, l2 = links[:2]
        w1 = wm.widget_for_node(n1)

        self.assertInWithCount(NodeEvent.OutputLinkAdded, w1._evt, 1)
        w1._evt.clear()
        workflow.remove_link(l1)

        self.assertInWithCount(NodeEvent.OutputLinkRemoved, w1._evt, 1)
        w3 = wm.widget_for_node(n3)
        w3._evt.clear()
        workflow.add_link(l1)
        self.assertInWithCount(NodeEvent.OutputLinkAdded, w1._evt, 1)
        self.assertInWithCount(NodeEvent.InputLinkAdded, w3._evt, 1)

        w1._evt.clear()
        workflow.set_runtime_env("tt", "aaa")
        self.assertInWithCount(NodeEvent.WorkflowEnvironmentChange, w1._evt, 1)

        w3._evt.clear()
        l1.set_runtime_state(SchemeLink.Pending)
        self.assertInWithCount(LinkEvent.InputLinkStateChange, w3._evt, 1)
        self.assertInWithCount(LinkEvent.OutputLinkStateChange, w1._evt, 1)

    def assertInWithCount(self, member, container, expected):
        counter = Counter(container)
        count = counter[member]
        if count != expected:
            msg = "Count of %s in %s is %i; expected %i" % (
                member, container, count, expected
            )
            self.fail(msg)

    def test_activation_on_delayed_creation_policy(self):
        workflow = self.scheme
        nodes = workflow.nodes
        wm = TestingWidgetManager()
        wm.set_creation_policy(WidgetManager.Normal)
        wm.set_workflow(workflow)
        n1, n2 = nodes[0], nodes[1]
        spy = QSignalSpy(wm.widget_for_node_added)
        QApplication.sendEvent(
            n1, WorkflowEvent(WorkflowEvent.NodeActivateRequest))
        self.assertEqual(len(spy), 1)
        self.assertIs(spy[0][0], n1)

        wm.set_creation_policy(WidgetManager.OnDemand)
        spy = QSignalSpy(wm.widget_for_node_added)
        QApplication.sendEvent(
            n2, WorkflowEvent(WorkflowEvent.NodeActivateRequest))
        self.assertEqual(len(spy), 1)
        self.assertIs(spy[0][0], n2)

    def test_actions(self):
        workflow = self.scheme
        nodes = workflow.nodes
        wm = TestingWidgetManager()
        wm.set_creation_policy(WidgetManager.Immediate)
        wm.set_workflow(workflow)
        w = wm.widget_for_node(nodes[0])
        w2 = wm.widget_for_node(nodes[2])
        espy = EventSpy(w2, QEvent.WindowActivate)
        ac = w.findChild(QAction, "action-canvas-raise-descendants")
        ac.trigger()
        if not espy.events():
            self.assertTrue(espy.wait(1000))
        self.assertTrue(w2.isActiveWindow())

        ac = w2.findChild(QAction, "action-canvas-raise-ancestors")
        espy = EventSpy(w, QEvent.Show)
        ac.trigger()
        if not espy.events():
            self.assertTrue(espy.wait(1000))
        self.assertTrue(w.isVisible())



from collections import OrderedDict
from xml.sax.saxutils import escape

from typing import Mapping, Callable

import docutils.core


def render_plain(content: str) -> str:
    """
    Return a html fragment for a plain pre-formatted text

    Parameters
    ----------
    content : str
        Plain text content

    Returns
    -------
    html : str
    """
    return '<p style="white-space: pre-wrap;">' + escape(content) + "</p>"


def render_html(content: str) -> str:
    """
    Return a html fragment unchanged.

    Parameters
    ----------
    content : str
        Html text.

    Returns
    -------
    html : str
    """
    return content


def render_markdown(content: str) -> str:
    """
    Return a html fragment from markdown text content

    Parameters
    ----------
    content : str
        A markdown formatted text

    Returns
    -------
    html : str
    """
    # commonmark >= 0.8.1; but only optionally. Many other packages may pin it
    # to <0.8 due to breaking changes.
    try:
        import commonmark
    except ImportError:
        return render_plain(content)
    else:
        return commonmark.commonmark(content)


def render_rst(content: str) -> str:
    """
    Return a html fragment from a RST text content

    Parameters
    ----------
    content : str
        A RST formatted text content

    Returns
    -------
    html : str
    """
    overrides = {
        "report_level": 10,  # suppress errors from appearing in the html
        "output-encoding": "utf-8"
    }
    html = docutils.core.publish_string(
        content, writer_name="html",
        settings_overrides=overrides
    )
    return html.decode("utf-8")


ContentRenderer = OrderedDict([
    ("text/plain", render_plain),
    ("text/rst", render_rst),
    ("text/x-rst", render_rst),
    ("text/markdown", render_markdown),
    ("text/html", render_html),
])  # type: Mapping[str, Callable[[str], str]]


def render_as_rich_text(content: str, content_type="text/plain") -> str:
    # split off the parameters (not supported)
    content_type, _, _ = content_type.partition(";")
    renderer = ContentRenderer.get(content_type.lower(), render_plain)
    try:
        return renderer(content)
    except (ImportError, ValueError):
        return render_plain(content)



import base64
import codecs

import unittest

from AnyQt.QtCore import QUrl

from orangecanvas.gui.test import QCoreAppTestCase

from orangecanvas.help.provider import sniff_html_charset, HtmlIndexProvider
from orangecanvas.registry import WidgetDescription
from orangecanvas.utils.asyncutils import get_event_loop
from orangecanvas.utils.shtools import temp_named_file


class TestUtils(unittest.TestCase):
    def test_sniff_html_charset(self):
        contents = (
            b'<html>\n'
            b' <header>\n'
            b'   <meta http-equiv="Content-Type" \n'
            b'         content="text/html; charset=cp1252" />\n'
            b' </header>\n'
            b'</html>'
        )
        self.assertEqual(sniff_html_charset(contents), "cp1252")
        self.assertEqual(sniff_html_charset(contents[:-7]), "cp1252")
        self.assertEqual(
            sniff_html_charset(contents[:-7] + b'.<>>,<<.\xfe\xff<'),
            "cp1252"
        )
        contents = (
            b'<html>\n'
            b' <header>\n'
            b'   <meta charset="utf-8" />\n'
            b' </header>\n'
            b'</html>'
        )
        self.assertEqual(sniff_html_charset(contents), "utf-8")
        self.assertEqual(sniff_html_charset(codecs.BOM_UTF8 + contents), "utf-8")

        self.assertEqual(sniff_html_charset(b''), None)
        self.assertEqual(sniff_html_charset(b'<html></html>'), None)

        self.assertEqual(
            sniff_html_charset(
                codecs.BOM_UTF16_BE +"<html></html>".encode("utf-16-be")
            ),
            'utf-16'
        )


def data_url(mimetype, payload):
    # type: (str, bytes) -> str
    payload = base64.b64encode(payload).decode("ascii")
    return "data:{};base64,{}".format(mimetype, payload)


class TestHtmlIndexProvider(QCoreAppTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.loop = get_event_loop()

    @classmethod
    def tearDownClass(cls):
        cls.loop.close()
        super().tearDownClass()

    def test(self):
        contents = (
            b'<html>\n'
            b' <header>\n'
            b'   <meta charset=cp1252" />\n'
            b' </header>\n'
            b' <body><%s id="widgets">\n'
            b'  <ul>\n'
            b'   <li><a href="a.html">aa</li>\n'
            b'  </ul>\n'
            b'  </div>\n'
            b'</html>'
        )
        for tag in (b"section", b"div"):
            with temp_named_file((contents % tag).decode("ascii"),) as fname:
                url = QUrl.fromLocalFile(fname)
                p = HtmlIndexProvider(url)
                loop = get_event_loop()
                desc = WidgetDescription(name="aa", id="aa", qualified_name="aa")
                res = loop.run_until_complete(p.search_async(desc))
                self.assertEqual(res, url.resolved(QUrl("a.html")))
                self.assertEqual(p.items, {"aa": "a.html"})



from typing import Any, Optional, Union

from AnyQt.QtWidgets import (
    QGraphicsItem, QGraphicsObject, QStyleOptionGraphicsItem, QWidget
)
from AnyQt.QtGui import (
    QPainterPath, QPainterPathStroker, QBrush, QPen, QPainter, QColor
)
from AnyQt.QtCore import Qt, QPointF, QRectF
from AnyQt.QtCore import pyqtSignal as Signal


class GraphicsPathObject(QGraphicsObject):
    """A QGraphicsObject subclass implementing an interface similar to
    QGraphicsPathItem, and also adding a positionChanged() signal
    """

    positionChanged = Signal([], ["QPointF"])

    def __init__(self, parent=None, **kwargs):
        # type: (Optional[QGraphicsItem], Any) -> None
        super().__init__(parent, **kwargs)
        self.setFlag(QGraphicsObject.ItemSendsGeometryChanges)

        self.__path = QPainterPath()
        self.__brush = QBrush(Qt.NoBrush)
        self.__pen = QPen()
        self.__boundingRect = None  # type: Optional[QRectF]

    def setPath(self, path):
        # type: (QPainterPath) -> None
        """Set the items `path` (:class:`QPainterPath`).
        """
        if self.__path != path:
            self.prepareGeometryChange()
            # Need to store a copy of object so the shape can't be mutated
            # without properly updating the geometry.
            self.__path = QPainterPath(path)
            self.__boundingRect = None
            self.update()

    def path(self):
        # type: () -> QPainterPath
        """Return the items path.
        """
        return QPainterPath(self.__path)

    def setBrush(self, brush):
        # type: (Union[QBrush, QColor, Qt.GlobalColor, Qt.BrushStyle]) -> None
        """Set the items `brush` (:class:`QBrush`)
        """
        if not isinstance(brush, QBrush):
            brush = QBrush(brush)

        if self.__brush != brush:
            self.__brush = QBrush(brush)
            self.update()

    def brush(self):
        # type: () -> QBrush
        """Return the items brush.
        """
        return QBrush(self.__brush)

    def setPen(self, pen):
        # type: (Union[QPen, QBrush, Qt.PenStyle]) -> None
        """Set the items outline `pen` (:class:`QPen`).
        """
        if not isinstance(pen, QPen):
            pen = QPen(pen)

        if self.__pen != pen:
            self.prepareGeometryChange()
            self.__pen = QPen(pen)
            self.__boundingRect = None
            self.update()

    def pen(self):
        # type: () -> QPen
        """Return the items pen.
        """
        return QPen(self.__pen)

    def paint(self, painter, option, widget=None):
        # type: (QPainter, QStyleOptionGraphicsItem, Optional[QWidget]) -> None
        if self.__path.isEmpty():
            return

        painter.save()
        painter.setPen(self.__pen)
        painter.setBrush(self.__brush)
        painter.drawPath(self.__path)
        painter.restore()

    def boundingRect(self):
        # type: () -> QRectF
        if self.__boundingRect is None:
            br = self.__path.controlPointRect()
            pen_w = self.__pen.widthF()
            self.__boundingRect = br.adjusted(-pen_w, -pen_w, pen_w, pen_w)

        return QRectF(self.__boundingRect)

    def shape(self):
        # type: () -> QPainterPath
        return shapeFromPath(self.__path, self.__pen)

    def itemChange(self, change, value):
        # type: (QGraphicsItem.GraphicsItemChange, Any) -> Any
        if change == QGraphicsObject.ItemPositionHasChanged:
            self.positionChanged.emit()
            self.positionChanged[QPointF].emit(value)

        return super().itemChange(change, value)


def shapeFromPath(path, pen):
    # type: (QPainterPath, QPen) -> QPainterPath
    """Create a QPainterPath shape from the `path` drawn with `pen`.
    """
    stroker = QPainterPathStroker()
    stroker.setCapStyle(pen.capStyle())
    stroker.setJoinStyle(pen.joinStyle())
    stroker.setMiterLimit(pen.miterLimit())
    stroker.setWidth(max(pen.widthF(), 1e-9))

    shape = stroker.createStroke(path)
    shape.addPath(path)

    return shape



"""

"""
import os
import string
import itertools
import logging
import urllib.parse
import warnings
from sysconfig import get_path

import typing
from typing import Dict, Optional, List, Tuple, Union, Callable, Sequence

import pkg_resources

from AnyQt.QtCore import QObject, QUrl, QDir

from ..utils.pkgmeta import get_dist_url, is_develop_egg
from . import provider

if typing.TYPE_CHECKING:
    from ..registry import WidgetRegistry, WidgetDescription
    Distribution = pkg_resources.Distribution
    EntryPoint = pkg_resources.EntryPoint

log = logging.getLogger(__name__)


class HelpManager(QObject):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._registry = None  # type: Optional[WidgetRegistry]
        self._providers = {}  # type: Dict[str, provider.HelpProvider]

    def set_registry(self, registry):
        # type: (Optional[WidgetRegistry]) -> None
        """
        Set the widget registry for which the manager should provide help.
        """
        if self._registry is not registry:
            self._registry = registry

    def registry(self):
        # type: () -> Optional[WidgetRegistry]
        """
        Return the previously set with set_registry.
        """
        return self._registry

    def initialize(self) -> None:
        warnings.warn(
            "`HelpManager.initialize` is deprecated and does nothing.",
            DeprecationWarning, stacklevel=2
        )
        return

    def get_provider(self, project: str) -> Optional[provider.HelpProvider]:
        """
        Return a `HelpProvider` for the `project` name.
        """
        provider = self._providers.get(project, None)
        if provider is None:
            try:
                dist = pkg_resources.get_distribution(project)
            except pkg_resources.ResolutionError:
                log.exception("Could not get distribution for '%s'", project)
            else:
                try:
                    provider = get_help_provider_for_distribution(dist)
                except Exception:  # noqa
                    log.exception("Error while initializing help "
                                  "provider for %r", project)

        if provider:
            self._providers[project] = provider
        return provider

    def get_help(self, url):
        # type: (QUrl) -> QUrl
        """
        """
        if url.scheme() == "help" and url.authority() == "search":
            return self.search(qurl_query_items(url))
        else:
            return url

    def description_by_id(self, desc_id):
        # type: (str) -> WidgetDescription
        reg = self._registry
        if reg is not None:
            return get_by_id(reg, desc_id)
        else:
            raise RuntimeError("No registry set. Cannot resolve")

    def search(self, query):
        # type: (Union[QUrl, Dict[str, str], Sequence[Tuple[str, str]]]) -> QUrl
        if isinstance(query, QUrl):
            query = qurl_query_items(query)

        query = dict(query)
        desc_id = query["id"]
        desc = self.description_by_id(desc_id)

        provider = None
        if desc.project_name:
            provider = self.get_provider(desc.project_name)

        if provider is not None:
            return provider.search(desc)
        else:
            raise KeyError(desc_id)

    async def search_async(self, query, timeout=2):
        if isinstance(query, QUrl):
            query = qurl_query_items(query)

        query = dict(query)
        desc_id = query["id"]
        desc = self.description_by_id(desc_id)

        provider = None
        if desc.project_name:
            provider = self.get_provider(desc.project_name)

        if provider is not None:
            return await provider.search_async(desc, timeout=timeout)
        else:
            raise KeyError(desc_id)


def get_by_id(registry, descriptor_id):
    # type: (WidgetRegistry, str) -> WidgetDescription
    for desc in registry.widgets():
        if desc.qualified_name == descriptor_id:
            return desc

    raise KeyError(descriptor_id)


def qurl_query_items(url: QUrl) -> List[Tuple[str, str]]:
    if not url.hasQuery():
        return []
    querystr = url.query()
    return urllib.parse.parse_qsl(querystr)


def _replacements_for_dist(dist):
    # type: (Distribution) -> Dict[str, str]
    replacements = {"PROJECT_NAME": dist.project_name,
                    "PROJECT_NAME_LOWER": dist.project_name.lower(),
                    "PROJECT_VERSION": dist.version,
                    "DATA_DIR": get_path("data")}
    try:
        replacements["URL"] = get_dist_url(dist)
    except KeyError:
        pass

    if is_develop_egg(dist):
        replacements["DEVELOP_ROOT"] = dist.location

    return replacements


def qurl_from_path(urlpath):
    # type: (str) -> QUrl
    if QDir(urlpath).isAbsolute():
        # deal with absolute paths including windows drive letters
        return QUrl.fromLocalFile(urlpath)
    return QUrl(urlpath, QUrl.TolerantMode)


def create_intersphinx_provider(entry_point):
    # type: (EntryPoint) -> Optional[provider.IntersphinxHelpProvider]
    locations = entry_point.resolve()
    if entry_point.dist is not None:
        replacements = _replacements_for_dist(entry_point.dist)
    else:
        replacements = {}

    formatter = string.Formatter()

    for target, inventory in locations:
        # Extract all format fields
        format_iter = formatter.parse(target)
        if inventory:
            format_iter = itertools.chain(format_iter,
                                          formatter.parse(inventory))
        # Names used in both target and inventory
        fields = {name for _, name, _, _ in format_iter if name}

        if not set(fields) <= set(replacements.keys()):
            continue

        target = formatter.format(target, **replacements)
        if inventory:
            inventory = formatter.format(inventory, **replacements)

        targeturl = qurl_from_path(target)
        if not targeturl.isValid():
            continue

        if targeturl.isLocalFile():
            if os.path.exists(os.path.join(target, "objects.inv")):
                inventory = QUrl.fromLocalFile(
                    os.path.join(target, "objects.inv"))
            else:
                log.info("Local doc root '%s' does not exist.", target)
                continue

        else:
            if not inventory:
                # Default inventory location
                inventory = targeturl.resolved(QUrl("objects.inv"))

        if inventory is not None:
            return provider.IntersphinxHelpProvider(
                inventory=inventory, target=target)
    return None


def create_html_provider(entry_point):
    # type: (EntryPoint) -> Optional[provider.SimpleHelpProvider]
    locations = entry_point.resolve()
    if entry_point.dist is not None:
        replacements = _replacements_for_dist(entry_point.dist)
    else:
        replacements = {}

    formatter = string.Formatter()

    for target in locations:
        # Extract all format fields
        format_iter = formatter.parse(target)
        fields = {name for _, name, _, _ in format_iter if name}

        if not set(fields) <= set(replacements.keys()):
            continue
        target = formatter.format(target, **replacements)

        targeturl = qurl_from_path(target)
        if not targeturl.isValid():
            continue

        if targeturl.isLocalFile():
            if not os.path.exists(target):
                log.info("Local doc root '%s' does not exist.", target)
                continue

        if target:
            return provider.SimpleHelpProvider(
                baseurl=QUrl.fromLocalFile(target))

    return None


def create_html_inventory_provider(entry_point):
    # type: (EntryPoint) -> Optional[provider.HtmlIndexProvider]
    locations = entry_point.resolve()
    if entry_point.dist is not None:
        replacements = _replacements_for_dist(entry_point.dist)
    else:
        replacements = {}

    formatter = string.Formatter()

    for target, xpathquery in locations:
        if isinstance(target, (tuple, list)):
            pass

        # Extract all format fields
        format_iter = formatter.parse(target)
        fields = {name for _, name, _, _ in format_iter if name}

        if not set(fields) <= set(replacements.keys()):
            continue
        target = formatter.format(target, **replacements)

        targeturl = qurl_from_path(target)
        if not targeturl.isValid():
            continue

        if targeturl.isLocalFile():
            if not os.path.exists(target):
                log.info("Local doc root '%s' does not exist", target)
                continue

            inventory = QUrl.fromLocalFile(target)
        else:
            inventory = QUrl(target)

        return provider.HtmlIndexProvider(
            inventory=inventory, xpathquery=xpathquery)

    return None


_providers = {
    "intersphinx": create_intersphinx_provider,
    "html-simple": create_html_provider,
    "html-index": create_html_inventory_provider,
}  # type: Dict[str, Callable[[EntryPoint], Optional[provider.HelpProvider]]]

_providers_cache = {}  # type: Dict[str, provider.HelpProvider]


def get_help_provider_for_distribution(dist):
    # type: (pkg_resources.Distribution) -> Optional[provider.HelpProvider]
    """
    Return a HelpProvider for the distribution.

    A 'orange.canvas.help' entry point is used to lookup one of the known
    provider classes, and the corresponding constructor factory is called
    with the entry point as the only parameter.

    Parameters
    ----------
    dist : Distribution

    Returns
    -------
    provider: Optional[provider.HelpProvider]
    """
    if dist.project_name in _providers_cache:
        return _providers_cache[dist.project_name]

    eps = dist.get_entry_map()
    entry_points = eps.get("orange.canvas.help", {})
    if not entry_points:
        # alternative name
        entry_points = eps.get("orangecanvas.help", {})

    provider = None
    for name, entry_point in entry_points.items():
        create = _providers.get(name, None)
        if create:
            try:
                provider = create(entry_point)
            except pkg_resources.DistributionNotFound as err:
                log.warning("Unsatisfied dependencies (%r)", err)
                continue
            except Exception as ex:
                log.exception("Exception {}".format(ex))
            if provider:
                log.info("Created %s provider for %s",
                         type(provider), dist)
                break

    if provider is not None:
        _providers_cache[dist.project_name] = provider
    return provider



"""
Orange Canvas Tool Dock widget

"""
import sys
import warnings
from typing import Optional, Any

from AnyQt.QtWidgets import (
    QWidget, QSplitter, QVBoxLayout, QAction, QSizePolicy, QApplication,
    QToolButton, QTreeView)
from AnyQt.QtGui import QPalette, QBrush, QDrag, QResizeEvent, QHideEvent

from AnyQt.QtCore import (
    Qt, QSize, QObject, QPropertyAnimation, QEvent, QRect, QPoint,
    QAbstractItemModel, QModelIndex, QPersistentModelIndex, QEventLoop,
    QMimeData
)
from AnyQt.QtCore import pyqtProperty as Property, pyqtSignal as Signal

from ..gui.toolgrid import ToolGrid
from ..gui.toolbar import DynamicResizeToolBar
from ..gui.quickhelp import QuickHelp
from ..gui.framelesswindow import FramelessWindow
from ..gui.utils import create_css_gradient, available_screen_geometry
from ..document.quickmenu import MenuPage
from .widgettoolbox import WidgetToolBox, iter_index, item_text, item_icon, item_tooltip
from ..registry.qt import QtWidgetRegistry


class SplitterResizer(QObject):
    """
    An object able to control the size of a widget in a QSplitter instance.
    """
    def __init__(self, parent=None, **kwargs):
        # type: (Optional[QObject], Any) -> None
        super().__init__(parent, **kwargs)
        self.__splitter = None  # type: Optional[QSplitter]
        self.__widget = None    # type: Optional[QWidget]
        self.__updateOnShow = True  # Need __update on next show event
        self.__animationEnabled = True
        self.__size = -1
        self.__expanded = False
        self.__animation = QPropertyAnimation(
            self, b"size_", self, duration=200
        )
        self.__action = QAction("toggle-expanded", self, checkable=True)
        self.__action.triggered[bool].connect(self.setExpanded)

    def setSize(self, size):
        # type: (int) -> None
        """
        Set the size of the controlled widget (either width or height
        depending on the orientation).

        .. note::
            The controlled widget's size is only updated when it it is shown.
        """
        if self.__size != size:
            self.__size = size
            self.__update()

    def size(self):
        # type: () -> int
        """
        Return the size of the widget in the splitter (either height of
        width) depending on the splitter orientation.
        """
        if self.__splitter and self.__widget:
            index = self.__splitter.indexOf(self.__widget)
            sizes = self.__splitter.sizes()
            return sizes[index]
        else:
            return -1

    size_ = Property(int, fget=size, fset=setSize)

    def setAnimationEnabled(self, enable):
        # type: (bool) -> None
        """Enable/disable animation."""
        self.__animation.setDuration(0 if enable else 200)

    def animationEnabled(self):
        # type: () -> bool
        return self.__animation.duration() == 0

    def setSplitterAndWidget(self, splitter, widget):
        # type: (QSplitter, QWidget) -> None
        """Set the QSplitter and QWidget instance the resizer should control.

        .. note:: the widget must be in the splitter.
        """
        if splitter and widget and not splitter.indexOf(widget) > 0:
            raise ValueError("Widget must be in a splitter.")

        if self.__widget is not None:
            self.__widget.removeEventFilter(self)
        if self.__splitter is not None:
            self.__splitter.removeEventFilter(self)

        self.__splitter = splitter
        self.__widget = widget

        if widget is not None:
            widget.installEventFilter(self)
        if splitter is not None:
            splitter.installEventFilter(self)

        self.__update()

        size = self.size()
        if self.__expanded and size == 0:
            self.open()
        elif not self.__expanded and size > 0:
            self.close()

    def toggleExpandedAction(self):
        # type: () -> QAction
        """Return a QAction that can be used to toggle expanded state.
        """
        return self.__action

    def toogleExpandedAction(self):
        warnings.warn(
            "'toogleExpandedAction is deprecated, use 'toggleExpandedAction' "
            "instead.", DeprecationWarning, stacklevel=2
        )
        return self.toggleExpandedAction()

    def open(self):
        # type: () -> None
        """Open the controlled widget (expand it to sizeHint).
        """
        self.__expanded = True
        self.__action.setChecked(True)

        if self.__splitter is None or self.__widget is None:
            return

        hint = self.__widget.sizeHint()

        if self.__splitter.orientation() == Qt.Vertical:
            end = hint.height()
        else:
            end = hint.width()

        self.__animation.setStartValue(0)
        self.__animation.setEndValue(end)
        self.__animation.start()

    def close(self):
        # type: () -> None
        """Close the controlled widget (shrink to size 0).
        """
        self.__expanded = False
        self.__action.setChecked(False)

        if self.__splitter is None or self.__widget is None:
            return

        self.__animation.setStartValue(self.size())
        self.__animation.setEndValue(0)
        self.__animation.start()

    def setExpanded(self, expanded):
        # type: (bool) -> None
        """Set the expanded state."""
        if self.__expanded != expanded:
            if expanded:
                self.open()
            else:
                self.close()

    def expanded(self):
        # type: () -> bool
        """Return the expanded state."""
        return self.__expanded

    def __update(self):
        # type: () -> None
        """Update the splitter sizes."""
        if self.__splitter and self.__widget:
            if sum(self.__splitter.sizes()) == 0:
                # schedule update on next show event
                self.__updateOnShow = True
                return

            splitter = self.__splitter
            index = splitter.indexOf(self.__widget)
            sizes = splitter.sizes()
            current = sizes[index]
            diff = current - self.__size
            sizes[index] = self.__size
            sizes[index - 1] = sizes[index - 1] + diff
            self.__splitter.setSizes(sizes)

    def eventFilter(self, obj, event):
        # type: (QObject, QEvent) -> bool
        if event.type() == QEvent.Resize and obj is self.__widget and \
                self.__animation.state() == QPropertyAnimation.Stopped:
            # Update the expanded state when the user opens/closes the widget
            # by dragging the splitter handle.
            assert self.__splitter is not None
            assert isinstance(event, QResizeEvent)
            if self.__splitter.orientation() == Qt.Vertical:
                size = event.size().height()
            else:
                size = event.size().width()

            if self.__expanded and size == 0:
                self.__action.setChecked(False)
                self.__expanded = False
            elif not self.__expanded and size > 0:
                self.__action.setChecked(True)
                self.__expanded = True

        if event.type() == QEvent.Show and obj is self.__splitter and \
                self.__updateOnShow:
            # Update the splitter state after receiving valid geometry
            self.__updateOnShow = False
            self.__update()
        return super().eventFilter(obj, event)


class QuickHelpWidget(QuickHelp):
    def minimumSizeHint(self):
        # type: () -> QSize
        """Reimplemented to allow the Splitter to resize the widget
        with a continuous animation.
        """
        hint = super().minimumSizeHint()
        return QSize(hint.width(), 0)


class CanvasToolDock(QWidget):
    """Canvas dock widget with widget toolbox, quick help and
    canvas actions.
    """
    def __init__(self, parent=None, **kwargs):
        # type: (Optional[QWidget], Any) -> None
        super().__init__(parent, **kwargs)

        self.__setupUi()

    def __setupUi(self):
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        self.toolbox = WidgetToolBox()

        self.help = QuickHelpWidget(objectName="quick-help")

        self.__splitter = QSplitter()
        self.__splitter.setOrientation(Qt.Vertical)

        self.__splitter.addWidget(self.toolbox)
        self.__splitter.addWidget(self.help)

        self.toolbar = DynamicResizeToolBar()
        self.toolbar.setMovable(False)
        self.toolbar.setFloatable(False)

        self.toolbar.setSizePolicy(QSizePolicy.Ignored,
                                   QSizePolicy.Preferred)

        layout.addWidget(self.__splitter, 10)
        layout.addWidget(self.toolbar)

        self.setLayout(layout)
        self.__splitterResizer = SplitterResizer(self)
        self.__splitterResizer.setSplitterAndWidget(self.__splitter, self.help)

    def setQuickHelpVisible(self, state):
        # type: (bool) -> None
        """Set the quick help box visibility status."""
        self.__splitterResizer.setExpanded(state)

    def quickHelpVisible(self):
        # type: () -> bool
        return self.__splitterResizer.expanded()

    def setQuickHelpAnimationEnabled(self, enabled):
        # type: (bool) -> None
        """Enable/disable the quick help animation."""
        self.__splitterResizer.setAnimationEnabled(enabled)

    def toggleQuickHelpAction(self):
        # type: () -> QAction
        """Return a checkable QAction for help show/hide."""
        return self.__splitterResizer.toggleExpandedAction()

    def toogleQuickHelpAction(self):
        warnings.warn(
            "'toogleQuickHelpAction' is deprecated, use "
            "'toggleQuickHelpAction' instead.", DeprecationWarning,
            stacklevel=2
        )
        return self.toggleQuickHelpAction()


class QuickCategoryToolbar(ToolGrid):
    """A toolbar with category buttons."""
    def __init__(self, parent=None, buttonSize=QSize(), iconSize=QSize(),
                 **kwargs):
        # type: (Optional[QWidget], QSize, QSize, Any) -> None
        super().__init__(parent, 1, buttonSize, iconSize,
                         Qt.ToolButtonIconOnly, **kwargs)
        self.__model = None  # type: Optional[QAbstractItemModel]

    def setColumnCount(self, count):
        raise Exception("Cannot set the column count on a Toolbar")

    def setModel(self, model):
        # type: (Optional[QAbstractItemModel]) -> None
        """
        Set the registry model.
        """
        if self.__model is not None:
            self.__model.dataChanged.disconnect(self.__on_dataChanged)
            self.__model.rowsInserted.disconnect(self.__on_rowsInserted)
            self.__model.rowsRemoved.disconnect(self.__on_rowsRemoved)
            self.clear()

        self.__model = model
        if model is not None:
            model.dataChanged.connect(self.__on_dataChanged)
            model.rowsInserted.connect(self.__on_rowsInserted)
            model.rowsRemoved.connect(self.__on_rowsRemoved)
            self.__initFromModel(model)

    def __initFromModel(self, model):
        # type: (QAbstractItemModel) -> None
        """
        Initialize the toolbar from the model.
        """
        for index in iter_index(model, QModelIndex()):
            action = self.createActionForItem(index)
            self.addAction(action)

    def createActionForItem(self, index):
        # type: (QModelIndex) -> QAction
        """
        Create the QAction instance for item at `index` (`QModelIndex`).
        """
        action = QAction(
            item_icon(index), item_text(index), self,
            toolTip=item_tooltip(index)
        )
        action.setData(QPersistentModelIndex(index))
        return action

    def createButtonForAction(self, action):
        # type: (QAction) -> QToolButton
        """
        Create a button for the action.
        """
        button = super().createButtonForAction(action)

        item = action.data()  # QPersistentModelIndex
        assert isinstance(item, QPersistentModelIndex)

        brush = item.data(Qt.BackgroundRole)
        if not isinstance(brush, QBrush):
            brush = item.data(QtWidgetRegistry.BACKGROUND_ROLE)
            if not isinstance(brush, QBrush):
                brush = self.palette().brush(QPalette.Button)

        palette = button.palette()
        palette.setColor(QPalette.Button, brush.color())
        palette.setColor(QPalette.Window, brush.color())
        button.setPalette(palette)
        button.setProperty("quick-category-toolbutton", True)

        style_sheet = ("QToolButton {\n"
                       "    background: %s;\n"
                       "    border: none;\n"
                       "    border-bottom: 1px solid palette(mid);\n"
                       "}")
        button.setStyleSheet(style_sheet % create_css_gradient(brush.color()))

        return button

    def __on_dataChanged(self, topLeft, bottomRight):
        # type: (QModelIndex, QModelIndex) -> None
        assert self.__model is not None
        parent = topLeft.parent()
        if not parent.isValid():
            for row in range(topLeft.row(), bottomRight.row() + 1):
                item = self.__model.index(row, 0)
                action = self.actions()[row]
                action.setText(item_text(item))
                action.setIcon(item_icon(item))
                action.setToolTip(item_tooltip(item))

    def __on_rowsInserted(self, parent, start, end):
        # type: (QModelIndex, int, int) -> None
        assert self.__model is not None
        if not parent.isValid():
            for row in range(start, end + 1):
                item = self.__model.index(row, 0)
                self.insertAction(row, self.createActionForItem(item))

    def __on_rowsRemoved(self, parent, start, end):
        # type: (QModelIndex, int, int) -> None
        assert self.__model is not None
        if not parent.isValid():
            for row in range(end, start - 1, -1):
                action = self.actions()[row]
                self.removeAction(action)


# This implements the (single category) node selection popup when the
# tooldock is not expanded.
class CategoryPopupMenu(FramelessWindow):
    """
    A menu popup from which nodes can be dragged or clicked/activated.
    """
    triggered = Signal(QAction)
    hovered = Signal(QAction)

    def __init__(self, parent=None, **kwargs):
        # type: (Optional[QWidget], Any) -> None
        super().__init__(parent, **kwargs)
        self.setWindowFlags(self.windowFlags() | Qt.Popup)

        layout = QVBoxLayout()
        layout.setContentsMargins(6, 6, 6, 6)

        self.__menu = MenuPage()
        self.__menu.setActionRole(QtWidgetRegistry.WIDGET_ACTION_ROLE)

        if sys.platform == "darwin":
            self.__menu.view().setAttribute(Qt.WA_MacShowFocusRect, False)

        self.__menu.triggered.connect(self.__onTriggered)
        self.__menu.hovered.connect(self.hovered)

        self.__dragListener = ItemViewDragStartEventListener(self)
        self.__dragListener.dragStarted.connect(self.__onDragStarted)

        self.__menu.view().viewport().installEventFilter(self.__dragListener)
        self.__menu.view().installEventFilter(self)

        layout.addWidget(self.__menu)

        self.setLayout(layout)

        self.__action = None  # type: Optional[QAction]
        self.__loop = None    # type: Optional[QEventLoop]

    def setCategoryItem(self, item):
        """
        Set the category root item (:class:`QStandardItem`).
        """
        warnings.warn(
            "setCategoryItem is deprecated. Use the more general 'setModel'"
            "and setRootIndex", DeprecationWarning, stacklevel=2
        )
        model = item.model()
        self.__menu.setModel(model)
        self.__menu.setRootIndex(item.index())

    def setModel(self, model):
        # type: (QAbstractItemModel) -> None
        """
        Set the model.

        Parameters
        ----------
        model : QAbstractItemModel
        """
        self.__menu.setModel(model)

    def setRootIndex(self, index):
        # type: (QModelIndex) -> None
        """
        Set the root index in `model`.

        Parameters
        ----------
        index : QModelIndex
        """
        self.__menu.setRootIndex(index)

    def setActionRole(self, role):
        # type: (Qt.ItemDataRole) -> None
        """
        Set the action role in model.

        This is an item role in `model` that returns a QAction for the item.

        Parameters
        ----------
        role : Qt.ItemDataRole
        """
        self.__menu.setActionRole(role)

    def popup(self, pos=None):
        # type: (Optional[QPoint]) -> None
        """
        Show the popup at `pos`.

        Parameters
        ----------
        pos : Optional[QPoint]
            The position in global screen coordinates
        """
        if pos is None:
            pos = self.pos()
        self.adjustSize()
        geom = widget_popup_geometry(pos, self)
        self.setGeometry(geom)
        self.show()
        self.__menu.view().setFocus()

    def exec(self, pos=None):
        # type: (Optional[QPoint]) -> Optional[QAction]
        self.popup(pos)
        self.__loop = QEventLoop()

        self.__action = None
        self.__loop.exec()
        self.__loop = None

        if self.__action is not None:
            action = self.__action
        else:
            action = None
        return action

    def exec_(self, *args, **kwargs):
        warnings.warn(
            "exec_ is deprecated, use exec", DeprecationWarning, stacklevel=2
        )
        return self.exec(*args, **kwargs)

    def hideEvent(self, event):
        # type: (QHideEvent) -> None
        if self.__loop is not None:
            self.__loop.exit(0)
        super().hideEvent(event)

    def __onTriggered(self, action):
        # type: (QAction) -> None
        self.__action = action
        self.triggered.emit(action)
        self.hide()

        if self.__loop:
            self.__loop.exit(0)

    def __onDragStarted(self, index):
        # type: (QModelIndex) -> None
        desc = index.data(QtWidgetRegistry.WIDGET_DESC_ROLE)
        icon = index.data(Qt.DecorationRole)

        drag_data = QMimeData()
        drag_data.setData(
            "application/vnd.orange-canvas.registry.qualified-name",
            desc.qualified_name.encode('utf-8')
        )
        drag = QDrag(self)
        drag.setPixmap(icon.pixmap(38))
        drag.setMimeData(drag_data)

        # TODO: Should animate (accept) hide.
        self.hide()

        # When a drag is started and the menu hidden the item's tool tip
        # can still show for a short time UNDER the cursor preventing a
        # drop.
        viewport = self.__menu.view().viewport()
        filter = ToolTipEventFilter()
        viewport.installEventFilter(filter)

        drag.exec(Qt.CopyAction)

        viewport.removeEventFilter(filter)

    def eventFilter(self, obj, event):
        if isinstance(obj, QTreeView) and event.type() == QEvent.KeyPress:
            key = event.key()
            if key in [Qt.Key_Return, Qt.Key_Enter]:
                curr = obj.currentIndex()
                if curr.isValid():
                    obj.activated.emit(curr)
                    return True
        return super().eventFilter(obj, event)


class ItemViewDragStartEventListener(QObject):
    dragStarted = Signal(QModelIndex)

    def __init__(self, parent=None, **kwargs):
        # type: (Optional[QObject], Any) -> None
        super().__init__(parent, **kwargs)
        self._pos = None    # type: Optional[QPoint]
        self._index = None  # type: Optional[QPersistentModelIndex]

    def eventFilter(self, viewport, event):
        # type: (QObject, QEvent) -> bool
        view = viewport.parent()

        if event.type() == QEvent.MouseButtonPress and \
                event.button() == Qt.LeftButton:

            index = view.indexAt(event.pos())

            if index is not None:
                self._pos = event.pos()
                self._index = QPersistentModelIndex(index)

        elif event.type() == QEvent.MouseMove and self._pos is not None and \
                ((self._pos - event.pos()).manhattanLength() >=
                 QApplication.startDragDistance()):
            assert self._index is not None
            if self._index.isValid():
                # Map to a QModelIndex in the model.
                index = QModelIndex(self._index)
                self._pos = None
                self._index = None

                self.dragStarted.emit(index)

        return super().eventFilter(view, event)


class ToolTipEventFilter(QObject):
    def eventFilter(self, receiver, event):
        # type: (QObject, QEvent) -> bool
        if event.type() == QEvent.ToolTip:
            return True

        return super().eventFilter(receiver, event)


def widget_popup_geometry(pos, widget):
    # type: (QPoint, QWidget) -> QRect
    widget.ensurePolished()

    if widget.testAttribute(Qt.WA_Resized):
        size = widget.size()
    else:
        size = widget.sizeHint()

    screen = QApplication.screenAt(pos)
    if screen is None:
        screen = QApplication.primaryScreen()
    screen_geom = screen.availableGeometry()
    size = size.boundedTo(screen_geom.size())
    geom = QRect(pos, size)

    if geom.top() < screen_geom.top():
        geom.moveTop(screen_geom.top())
    if geom.left() < screen_geom.left():
        geom.moveLeft(screen_geom.left())

    bottom_margin = screen_geom.bottom() - geom.bottom()
    right_margin = screen_geom.right() - geom.right()
    if bottom_margin < 0:
        # Falls over the bottom of the screen, move it up.
        geom.translate(0, bottom_margin)

    # TODO: right to left locale
    if right_margin < 0:
        # Falls over the right screen edge, move the menu to the
        # other side of pos.
        geom.translate(-size.width(), 0)

    return geom


def popup_position_from_source(popup, source, orientation=Qt.Vertical):
    # type: (QWidget, QWidget, Qt.Orientation) -> QPoint
    popup.ensurePolished()
    source.ensurePolished()

    if popup.testAttribute(Qt.WA_Resized):
        size = popup.size()
    else:
        size = popup.sizeHint()

    screen_geom = available_screen_geometry(source)
    source_rect = QRect(source.mapToGlobal(QPoint(0, 0)), source.size())

    if orientation == Qt.Vertical:
        if source_rect.right() + size.width() < screen_geom.right():
            x = source_rect.right()
        else:
            x = source_rect.left() - size.width()

        # bottom overflow
        dy = source_rect.top() + size.height() - screen_geom.bottom()
        if dy < 0:
            y = source_rect.top()
        else:
            y = max(screen_geom.top(), source_rect.top() - dy)
    else:
        # right overflow
        dx = source_rect.left() + size.width() - screen_geom.right()
        if dx < 0:
            x = source_rect.left()
        else:
            x = max(source_rect.left() - dx, screen_geom.left())

        if source_rect.bottom() + size.height() < screen_geom.bottom():
            y = source_rect.bottom()
        else:
            y = source_rect.top() - size.height()

    return QPoint(x, y)



"""
Tests for items
"""
from AnyQt.QtWidgets import QGraphicsScene, QGraphicsView
from AnyQt.QtGui import QPainter

from orangecanvas.gui.test import QAppTestCase


class TestItems(QAppTestCase):
    def setUp(self):
        super().setUp()

        self.scene = QGraphicsScene()
        self.view = QGraphicsView(self.scene)
        self.view.setRenderHints(
            QPainter.Antialiasing |
            QPainter.SmoothPixmapTransform |
            QPainter.TextAntialiasing
        )
        self.view.resize(500, 300)
        self.view.show()

    def tearDown(self):
        self.scene.clear()
        self.scene.deleteLater()
        self.view.deleteLater()
        del self.scene
        del self.view
        super().tearDown()



"""
Tests for settings utility module.

"""
import tempfile

import unittest
import logging

from AnyQt.QtCore import QSettings
from ..settings import Settings, config_slot, QSettings_readArray, \
    QSettings_writeArray, QSettings_writeArrayItem
from ...gui import test


class TestUserSettings(test.QAppTestCase):
    def setUp(self):
        logging.basicConfig()
        super().setUp()

    def test_settings(self):
        spec = [config_slot("foo", bool, True, "foo doc"),
                config_slot("bar", int, 0, "bar doc"),
                ]

        store = QSettings(QSettings.IniFormat, QSettings.UserScope,
                          "biolab.si", "Orange Canvas Unit Tests")
        store.clear()
        settings = Settings(defaults=spec, store=store)

        self.assertEqual(settings["foo"], True)
        self.assertEqual(settings.get("foo"), True)

        self.assertEqual(settings.get("bar", 3), 0, "Defaults")
        self.assertEqual(settings.get("bar"), settings["bar"])

        self.assertEqual(settings.get("does not exist", "^&"), "^&",
                         "get with default")

        self.assertIs(settings.get("does not exist"), None)

        with self.assertRaises(KeyError):
            settings["does not exist"]

        self.assertTrue(settings.isdefault("foo"))

        changed = []

        settings.valueChanged.connect(
            lambda key, value: changed.append((key, value))
        )

        settings["foo"] = False
        self.assertEqual(changed[-1], ("foo", False), "valueChanged signal")
        self.assertEqual(len(changed), 1)
        self.assertEqual(settings["foo"], False, "updated value")
        self.assertEqual(settings.get("foo"), False)
        self.assertFalse(settings.isdefault("foo"))

        settings["bar"] = 1
        self.assertEqual(changed[-1], ("bar", 1), "valueChanged signal")
        self.assertEqual(len(changed), 2)
        self.assertEqual(settings["bar"], 1)
        self.assertFalse(settings.isdefault("bar"))

        del settings["bar"]
        self.assertEqual(settings["bar"], 0)
        self.assertEqual(changed[-1], ("bar", 0))

        # Only str or unicode can be keys
        with self.assertRaises(TypeError):
            settings[1] = 3

        # value type check
        with self.assertRaises(TypeError):
            settings["foo"] = []

        self.assertEqual(len(changed), 3)

        settings.add_default_slot(config_slot("foobar/foo", object, None, ""))

        group = settings.group("foobar")

        self.assertIs(group["foo"], None)
        group["foo"] = 3

        self.assertEqual(changed[-1], ("foobar/foo", 3))

        group["foonew"] = 5
        self.assertIn("foobar/foonew", settings)

        settings["newkey"] = "newkeyvalue"
        self.assertIn("newkey", settings)

        group1 = group.group("bar")
        group1["barval"] = "barval"

        self.assertIn("foobar/bar/barval", settings)

        settings["foobar/bar/barval"] = 5
        self.assertEqual(changed[-1], ("foobar/bar/barval", 5))

        settings.clear()
        self.assertSetEqual(set(settings.keys()),
                            set(["foo", "bar", "foobar/foo"]))


class TestQSettings_array(unittest.TestCase):
    filename = ""  # type: str

    def setUp(self):
        self.file = tempfile.NamedTemporaryFile()
        self.filename = self.file.name
        self.settings = QSettings(self.filename, QSettings.IniFormat)

    def tearDown(self):
        self.settings.sync()
        del self.settings
        self.file.close()

    def test_readwrite_array(self):
        s = self.settings
        scheme = {
            "name": str,
            "price": int
        }
        items = QSettings_readArray(s, "items", scheme)
        self.assertSequenceEqual(items, [])
        items_ = [
            {"name": "apple", "price": 10},
            {"name": "pear", "price": 12},
        ]
        QSettings_writeArray(s, "items", items_)
        items = QSettings_readArray(s, "items", scheme)
        self.assertSequenceEqual(items, items_)
        scheme = {
            "quality": (int, -1),
            **scheme
        }
        items = QSettings_readArray(s, "items", scheme)
        self.assertSequenceEqual(items, [{"quality": -1, **d} for d in items_])

        QSettings_writeArrayItem(
            s, "items", 1, {"name": "banana", "price": 5, "quality": 5},
            arraysize=2
        )
        items = QSettings_readArray(s, "items", scheme)
        self.assertSequenceEqual(items, [
            {"name": "apple", "price": 10, "quality": -1},
            {"name": "banana", "price": 5, "quality": 5}
        ])



import logging
import unittest
from contextlib import contextmanager
from functools import wraps
from typing import Iterable
from unittest.mock import patch, Mock

from orangecanvas import config
from orangecanvas.application.canvasmain import CanvasMainWindow
from orangecanvas.config import Config, EntryPoint
from orangecanvas.gui.test import QAppTestCase
from orangecanvas.main import Main
from orangecanvas.registry import WidgetDiscovery
from orangecanvas.registry.tests import set_up_modules, tear_down_modules
from orangecanvas.scheme import Scheme
from orangecanvas.utils.shtools import temp_named_file


class TestMain(unittest.TestCase):
    def test_params(self):
        m = Main()
        m.parse_arguments(["-", "--config", "foo.bar", "that"])
        self.assertEqual(m.arguments, ["that"])
        self.assertEqual(m.options.config, "foo.bar")
        m = Main()
        m.parse_arguments(["-", "-l3"])
        self.assertEqual(m.options.log_level, logging.WARNING)
        m = Main()
        m.parse_arguments(["-", "-l", "warn"])
        self.assertEqual(m.options.log_level, logging.WARNING)

    def test_style_param_compat(self):
        # test old '--style' parameter handling
        m = Main()
        m.parse_arguments(["-", "--style", "windows"])
        self.assertEqual(m.arguments, ["-style", "windows"])

        m = Main()
        m.parse_arguments(["-", "--qt", "-stylesheet path.qss"])
        self.assertEqual(m.arguments, ["-stylesheet", "path.qss"])

    def test_main_argument_parser(self):
        class Main2(Main):
            def argument_parser(self):
                p = super().argument_parser()
                p.add_argument("--foo", type=str, default=None)
                return p
        m = Main2()
        m.parse_arguments(["-", "-l", "warn", "--foo", "bar"])
        self.assertEqual(m.options.foo, "bar")


@contextmanager
def patch_main_application(app):
    def setup_application(self: Main):
        self.application = app
    with patch.object(Main, "setup_application",  setup_application):
        yield


def with_patched_main_application(f):
    @wraps(f)
    def wrapped(self: QAppTestCase, *args, **kwargs):
        with patch_main_application(self.app):
            return f(self, *args, **kwargs)
    return wrapped


class TestConfig(Config):
    def init(self):
        return

    def widget_discovery(self, *args, **kwargs):
        return WidgetDiscovery(*args, **kwargs)

    def widgets_entry_points(self):  # type: () -> Iterable[EntryPoint]
        pkg = "orangecanvas.registry.tests"
        return (
            EntryPoint.parse(f"add = {pkg}.operators.add"),
            EntryPoint.parse(f"sub = {pkg}.operators.sub")
        )

    def workflow_constructor(self, *args, **kwargs):
        return Scheme(*args, **kwargs)


class TestMainGuiCase(QAppTestCase):
    def setUp(self):
        super().setUp()
        self.app.fileOpenRequest = Mock()
        self._config = config.default
        set_up_modules()

    def tearDown(self):
        tear_down_modules()
        config.default = self._config
        del self.app.fileOpenRequest
        del self._config
        super().tearDown()

    @with_patched_main_application
    def test_main_show_splash_screen(self):
        m = Main()
        m.parse_arguments(["-", "--config", f"{__name__}.TestConfig"])
        m.activate_default_config()
        m.show_splash_message("aa")
        m.close_splash_screen()

    @with_patched_main_application
    def test_discovery(self):
        m = Main()
        m.parse_arguments(["-", "--config", f"{__name__}.TestConfig"])
        m.activate_default_config()
        m.run_discovery()
        self.assertTrue(bool(m.registry.widgets()))
        self.assertTrue(bool(m.registry.categories()))

    @with_patched_main_application
    def test_run(self):
        m = Main()
        with patch.object(self.app, "exec", lambda: 42):
            res = m.run(["-", "--no-welcome", "--no-splash"])
        self.assertEqual(res, 42)

    @with_patched_main_application
    def test_run(self):
        m = Main()
        with patch.object(self.app, "exec", lambda: 42), \
             patch.object(CanvasMainWindow, "open_scheme_file", Mock()), \
             temp_named_file('<scheme version="2.0"></scheme>') as fname:
            res = m.run(["-", "--no-welcome", "--no-splash", fname])
            CanvasMainWindow.open_scheme_file.assert_called_with(fname)
        self.assertEqual(res, 42)



"""
Test widget discovery

"""
import logging

import unittest

from ..discovery import WidgetDiscovery, widget_descriptions_from_package
from ..description import CategoryDescription, WidgetDescription
from ..utils import category_from_package_globals


class TestDiscovery(unittest.TestCase):

    def setUp(self):
        logging.basicConfig()
        from . import set_up_modules, operators, constants
        set_up_modules()
        self.operators = operators
        self.constants = constants

    def tearDown(self):
        from . import tear_down_modules
        tear_down_modules()

    def discovery_class(self):
        return WidgetDiscovery()

    def test_handle(self):
        disc = self.discovery_class()

        desc = CategoryDescription(name="C", qualified_name="M.C")
        disc.handle_category(desc)

        desc = WidgetDescription(name="SomeWidget", id="some.widget",
                                 qualified_name="Some.Widget",
                                 category="C",)
        disc.handle_widget(desc)

    def test_process_module(self):
        disc = self.discovery_class()
        disc.process_category_package(self.operators.__name__)
        disc.process_widget_module(self.constants.one.__name__)

    def test_process_loader(self):
        disc = self.discovery_class()

        def callable(discovery):
            desc = CategoryDescription(
                name="Data", qualified_name="Data")

            discovery.handle_category(desc)

            desc = WidgetDescription(
                name="CSV", id="some.id", qualified_name="Some.widget",
                inputs=[], category="Data",
            )
            discovery.handle_widget(desc)

        disc.process_loader(callable)

    def test_process_iter(self):
        disc = self.discovery_class()
        cat_desc = category_from_package_globals(
            self.operators.__name__,
        )
        # TODO: Fix (the widget_description_package does not iterate
        # over faked package (no valid operator.__path__)
        wid_desc = widget_descriptions_from_package(
            self.operators.__name__,
        )
        disc.process_iter([cat_desc] + wid_desc)

    def test_run(self):
        disc = self.discovery_class()
        disc.run("example.does.not.exist.but.it.does.not.matter.")



"""
Tests for SchemeLink
"""
import warnings

from ...gui import test
from ...registry.tests import small_testing_registry

from ...registry.description import InputSignal, OutputSignal, Dynamic

from .. import SchemeNode, SchemeLink, IncompatibleChannelTypeError
from ..link import resolved_valid_types, _classify_connection


class A:
    pass


class B(A):
    pass


class C:
    pass


class TestSchemeLink(test.QAppTestCase):
    def test_link(self):
        reg = small_testing_registry()
        one_desc = reg.widget("one")
        add_desc = reg.widget("add")
        unit_desc = reg.widget("unit")

        one_node = SchemeNode(one_desc)
        add_node = SchemeNode(add_desc)
        unit_node = SchemeNode(unit_desc)

        link1 = SchemeLink(one_node, one_node.output_channel("value"),
                           add_node,
                           add_node.input_channel("left"))

        self.assertEqual(link1.source_types(), (int, ))
        self.assertEqual(link1.sink_types(), (int, ))

        with self.assertRaises(ValueError):
            SchemeLink(add_node, "right", one_node, "$$$[")

        with self.assertRaises(IncompatibleChannelTypeError):
            SchemeLink(unit_node, "value", add_node, "right")

    def test_utils(self):
        with warnings.catch_warnings(record=True):
            self.assertTupleEqual(
                resolved_valid_types(("int", __name__ + ".NoSuchType", "str")),
                (int, str),
            )

        source = OutputSignal("A", (A,))
        source_d = OutputSignal("A", (A,), flags=Dynamic)
        source_ac_d = OutputSignal("A", (A, C), flags=Dynamic)
        source_ac = OutputSignal("A", (B, C))
        sink_a = InputSignal("A", (A,), "a")
        sink_b = InputSignal("B", (B,), "b")
        sink_c = InputSignal("C", (C,), "c")
        sink_bc = InputSignal("C", (B, C,), "c")

        t1, t2 = _classify_connection(source, sink_a)
        self.assertTrue(t1)
        self.assertFalse(t2)

        t1, t2 = _classify_connection(source, sink_b)
        self.assertFalse(t1)
        self.assertFalse(t2)

        t1, t2 = _classify_connection(source_d, sink_b)
        self.assertFalse(t1)
        self.assertTrue(t2)

        t1, t2 = _classify_connection(source_d, sink_a)
        self.assertTrue(t1)
        self.assertTrue(t2)

        t1, t2 = _classify_connection(source_d, sink_c)
        self.assertFalse(t1)
        self.assertFalse(t2)

        t1, t2 = _classify_connection(source_d, sink_bc)
        self.assertFalse(t1)
        self.assertTrue(t2)

        t1, t2 = _classify_connection(source_ac_d, sink_bc)
        self.assertFalse(t1)
        self.assertTrue(t2)

        t1, t2 = _classify_connection(source_ac, sink_bc)
        self.assertTrue(t1)
        self.assertFalse(t2)



import weakref
from types import SimpleNamespace as namespace
from typing import Optional, TypeVar, Generic

from AnyQt.QtCore import QObject, Qt

__all__ = [
    "qobjref", "qobjref_weak"
]

Q = TypeVar("Q", bound=QObject)


class qobjref(Generic[Q]):
    """
    A 'guarded reference' to a QObject.

    An instance of a `qobjref` holds a reference to a `QObject` for as long
    as it is alive (not destroyed by C++ destructor).

    Example
    -------
    >>> import sip
    >>> obj = QObject()
    >>> ref = qobjref(obj)
    >>> assert ref() is obj
    >>> sip.delete(obj)  # forcibly delete it
    >>> assert ref() is None

    Note
    ----
    This is not thread safe in the sense that the object can be deleted
    from a different thread while the ref() is returning it.

    Note
    ----
    `qobjref` keeps a reference to the object, meaning it will not be
    garbage collected until qobjref is also reaped. Use qobjref_weak
    for weak references.

    See also
    --------
    QPointer
    """
    __slots__ = ("__obj", "__state", "__weakref__")

    def __init__(self, obj):
        # type: (Q) -> None
        assert isinstance(obj, QObject)
        self.__obj = obj

        def finalize(_weakref):
            # finalize when self is collected
            if state.objref is None:
                return
            objref = state.objref()
            if objref is not None:
                objref.destroyed.disconnect(state.zero_ref)

        # destroy/zero the reference when QObject is destroyed
        def zero_ref():
            selfref = state.selfref()
            if selfref is not None:
                selfref.__obj = None
            state.objref = None

        state = namespace(
            selfref=weakref.ref(self, finalize),
            objref=weakref.ref(obj),
            zero_ref=zero_ref,
            finalize=finalize)
        self.__state = state

        # Must not capture self in the zero_ref's closure
        obj.destroyed.connect(zero_ref, Qt.DirectConnection)

    def __call__(self):
        # type: () -> Optional[Q]
        """
        Return the QObject instance or None if it was destroyed.

        Return
        ------
        obj : Optional[QObject]
        """
        return self.__obj

    def __repr__(self):
        objrep = "to " + repr(self.__obj)
        if self.__obj is None:
            objrep = "dead"

        return "<qobjref at 0x%x; " % id(self) + objrep + ">"


class qobjref_weak(Generic[Q]):
    """
    A weak 'guarded reference' to a QObject.

    Similar to `qobjref`, except that the reference to the QObject is weak

    Example
    -------
    >>> import sip
    >>> obj = QObject()
    >>> ref = qobjref_weak(obj)
    >>> assert ref() is obj
    >>> sip.delete(obj)  # forcibly delete it
    >>> assert ref() is None
    >>> obj = QObject()
    >>> ref = qobjref_weak(obj)
    >>> assert ref() is obj
    >>> del obj  # assuming ref count is 1
    >>> assert ref() is None

    Note
    ----
    This is not thread safe in the sense that the object can be
    deleted from a different thread while the ref() is returning it.

    See also
    --------
    qobjref, QPointer
    """
    __slots__ = ("__obj_ref", "__state", "__weakref__")

    def __init__(self, obj):
        # type: (Q) -> None
        assert isinstance(obj, QObject)
        self.__obj_ref = weakref.ref(obj)

        # finalize when self is collected
        def disconnect(_weakref):
            if state.objref is None:
                return
            objref = state.objref()
            if objref is not None:
                objref.destroyed.disconnect(state.zero_ref)

        # destroy/zero the reference when QObject is destroyed
        def zero_ref():
            selfref = state.selfref()
            if selfref is not None:
                selfref.__obj_ref = None
            state.objref = None

        state = namespace(
            selfref=weakref.ref(self, disconnect),
            objref=weakref.ref(obj),
            zero_ref=zero_ref,
            disconnect=disconnect)
        self.__state = state

        # Must not capture self in the zero_ref's closure
        obj.destroyed.connect(zero_ref, Qt.DirectConnection)

    def __call__(self):
        # type: () -> Optional[Q]
        """
        Return the QObject instance or None if it is no longer alive

        Return
        ------
        obj : Optional[QObject]
        """
        ref = self.__obj_ref
        if ref is not None:
            return ref()
        else:
            return None

    def __repr__(self):
        obj = self.__call__()
        if obj is not None:
            objrep = "to " + repr(obj)
        else:
            objrep = "dead"
        return "<qobjref_weak at 0x%x; " % id(self) + objrep + ">"



"""
Test for canvas toolbox.
"""

from AnyQt.QtWidgets import (
    QWidget, QToolBar, QTextEdit, QSplitter, QApplication
)
from AnyQt.QtCore import Qt, QTimer, QPoint

from ...registry import tests as registry_tests
from ...registry.qt import QtWidgetRegistry
from ...gui.dock import CollapsibleDockWidget

from ..canvastooldock import (
    WidgetToolBox, CanvasToolDock, SplitterResizer, QuickCategoryToolbar,
    CategoryPopupMenu, popup_position_from_source, widget_popup_geometry
)

from ...gui import test


class TestCanvasDockWidget(test.QAppTestCase):
    def test_dock(self):
        reg = registry_tests.small_testing_registry()
        reg = QtWidgetRegistry(reg, parent=self.app)

        toolbox = WidgetToolBox()
        toolbox.setObjectName("widgets-toolbox")
        toolbox.setModel(reg.model())

        text = QTextEdit()
        splitter = QSplitter()
        splitter.setOrientation(Qt.Vertical)

        splitter.addWidget(toolbox)
        splitter.addWidget(text)

        dock = CollapsibleDockWidget()
        dock.setExpandedWidget(splitter)

        toolbar = QToolBar()
        toolbar.addAction("1")
        toolbar.setOrientation(Qt.Vertical)
        toolbar.setMovable(False)
        toolbar.setFloatable(False)
        dock.setCollapsedWidget(toolbar)

        dock.show()
        self.qWait()

    def test_canvas_tool_dock(self):
        reg = registry_tests.small_testing_registry()
        reg = QtWidgetRegistry(reg, parent=self.app)

        dock = CanvasToolDock()
        dock.toolbox.setModel(reg.model())

        dock.show()
        self.qWait()

    def test_splitter_resizer(self):
        w = QSplitter(orientation=Qt.Vertical)
        w.addWidget(QWidget())
        text = QTextEdit()
        w.addWidget(text)
        resizer = SplitterResizer(parent=None)
        resizer.setSplitterAndWidget(w, text)

        def toogle():
            if resizer.size() == 0:
                resizer.open()
            else:
                resizer.close()

        w.show()
        timer = QTimer(resizer, interval=100)
        timer.timeout.connect(toogle)
        timer.start()
        toogle()
        self.qWait()
        timer.stop()

    def test_category_toolbar(self):
        reg = registry_tests.small_testing_registry()
        reg = QtWidgetRegistry(reg, parent=self.app)

        w = QuickCategoryToolbar()
        w.setModel(reg.model())
        w.show()
        self.qWait()


class TestPopupMenu(test.QAppTestCase):
    def test(self):
        reg = registry_tests.small_testing_registry()
        reg = QtWidgetRegistry(reg, parent=self.app)
        model = reg.model()

        w = CategoryPopupMenu()
        w.setModel(model)
        w.setRootIndex(model.index(0, 0))
        w.popup()
        self.qWait()

    def test_popup_position(self):
        popup = CategoryPopupMenu()
        screen = popup.screen()
        screen_geom = screen.availableGeometry()
        popup.setMinimumHeight(screen_geom.height() + 20)
        w = QWidget()
        w.setGeometry(
            screen_geom.left() + 100, screen_geom.top() + 100, 20, 20
        )
        pos = popup_position_from_source(popup, w)
        self.assertTrue(screen_geom.contains(pos))
        pos = QPoint(screen_geom.top() - 100, screen_geom.left() - 100)
        geom = widget_popup_geometry(pos, popup)
        self.assertEqual(screen_geom.intersected(geom), geom)
